#include <chronoptics/tof.hpp>
#include <iostream>
#include <thread>

using namespace chronoptics;

// This is how PointCloudLibrary represents XYZ
struct PointXYZ {
  float x;
  float y;
  float z;
  float pack;
};

void use_point_cloud(std::vector<PointXYZ> &) {
  std::cout << "4. Using point cloud\n";
  // Some expensive operation
  std::this_thread::sleep_for(std::chrono::milliseconds{100});
}

void simple(tof::Camera &cam, size_t size) {
  std::cout << "Simple usage\n\n";

  // User pointer supplied needs to live up to the point the callback function
  // is called.
  std::vector<PointXYZ> point_cloud(size);

  // Same holds for the function
  std::function<void(uint8_t *, size_t)> point_cloud_no_longer_used;

  // One option is to re add the user pointer
  point_cloud_no_longer_used = [&](uint8_t *ptr, size_t capacity) {
    if (cam.is_streaming()) {
      cam.add_user_pointer(ptr, capacity, point_cloud_no_longer_used,
                           tof::FrameType::XYZ);
      std::cout << "3. Re added point cloud\n";
    } else {
      std::cout << "2. Stopped re adding point cloud\n";
    }
  };

  cam.add_user_pointer(reinterpret_cast<uint8_t *>(point_cloud.data()),
                       point_cloud.size() * sizeof(PointXYZ),
                       point_cloud_no_longer_used, tof::FrameType::XYZ);

  cam.start();

  for (size_t i = 0; i < 20; i++) {
    std::cout << "1. Received frames\n";
    auto frames = cam.get_frames();

    // Because of the expensive operation we will drop frames, so not every
    // frame will be user allocated
    for (auto &frame : frames) {
      if (frame.user_allocated()) {
        // This is the frame that uses point_cloud
        use_point_cloud(point_cloud);
      }
    }

    // When the frames get destructed the destruct callback should be called
  }

  cam.stop();

  cam.clear_user_pointers();
}

void advanced(tof::Camera &cam, size_t size) {
  std::cout << "\n\nAdvanced usage:\n\n";

  // Here we go with a different tack, since the point_cloud and function need
  // to live up to the callback call we allocate it on the heap in this struct
  // and use it to manage lifetime
  struct CallbackData {
    CallbackData(tof::Camera &cam_, size_t size)
        : cam(cam_), point_cloud(size) {}

    tof::Camera &cam;
    std::vector<PointXYZ> point_cloud;
    bool finished{false};

    tof_user_pointer_destructed_t cb = [](uint8_t *, size_t, void *user_data) {
      auto data = static_cast<CallbackData *>(user_data);
      data->clean_up();
    };

    void add_user_pointer() {
      cam.add_user_pointer(reinterpret_cast<uint8_t *>(point_cloud.data()),
                           point_cloud.size() * sizeof(PointXYZ), cb, this,
                           tof::FrameType::XYZ);
    }

    void clean_up() {
      if (finished || !cam.is_streaming()) {
        std::cout << "2. Deleted point cloud\n";
        delete this;
      } else {
        std::cout << "3. Readded lost point cloud\n";
        add_user_pointer();
      }
    }
  };

  const size_t user_pointer_capacity = 10;
  cam.set_user_pointer_capacity(user_pointer_capacity);

  for (size_t i = 0; i < user_pointer_capacity; i++) {
    // Fill user pointers up to capacity, an error will be thrown if you
    // over fill
    auto data = new CallbackData(cam, size);
    data->add_user_pointer();
  }

  cam.start();

  for (size_t i = 0; i < 20; i++) {
    std::cout << "1. Received frames\n";
    auto frames = cam.get_frames();

    for (auto &frame : frames) {
      if (frame.user_allocated()) {
        auto data = reinterpret_cast<CallbackData *>(frame.user_pointer());
        use_point_cloud(data->point_cloud);
        data->finished = true;

        // Add a new user pointer since the one here will soon be deleted, you
        // could also cycle through to save heap allocations
        auto new_data = new CallbackData(cam, size);
        new_data->add_user_pointer();
      }
    }
  }

  cam.stop();

  cam.clear_user_pointers();
}

int main(int argc, char **argv) {
  if (argc != 2) {
    std::cout << "Usage: user_pointer <serial>\n";
    return -1;
  }

  tof::KeaCamera cam(tof::ProcessingConfig{}, argv[1]);
  tof::select_streams(cam, {tof::FrameType::XYZ});

  // Figure out Data size
  auto roi = cam.get_camera_config().get_roi(0);
  const size_t size = roi.get_img_rows() * roi.get_img_cols();

  simple(cam, size);
  advanced(cam, size);

  return 0;
}